BEGIN {				# Magic Perl CORE pragma
    if ($ENV{PERL_CORE}) {
        chdir 't' if -d 't';
        @INC = '../lib';
    }
}

use Test::More tests => 2;

BEGIN { use_ok('threads') }
BEGIN { use_ok('threads::farm') }
